/* GENERATED FILE */
/* use ./gendef0 to regenerate this file */
#ifndef ENOSYS
#define ENOSYS 0
#endif
#ifndef EXDEV
#define EXDEV 0
#endif
#ifndef EBADR
#define EBADR 0
#endif
#ifndef SEEK_DATA
#define SEEK_DATA 0
#endif
#ifndef SEEK_HOLE
#define SEEK_HOLE 0
#endif
#ifndef POSIX_FADV_NORMAL
#define POSIX_FADV_NORMAL 0
#endif
#ifndef POSIX_FADV_SEQUENTIAL
#define POSIX_FADV_SEQUENTIAL 0
#endif
#ifndef POSIX_FADV_RANDOM
#define POSIX_FADV_RANDOM 0
#endif
#ifndef POSIX_FADV_NOREUSE
#define POSIX_FADV_NOREUSE 0
#endif
#ifndef POSIX_FADV_WILLNEED
#define POSIX_FADV_WILLNEED 0
#endif
#ifndef POSIX_FADV_DONTNEED
#define POSIX_FADV_DONTNEED 0
#endif
#ifndef POSIX_MADV_NORMAL
#define POSIX_MADV_NORMAL 0
#endif
#ifndef POSIX_MADV_SEQUENTIAL
#define POSIX_MADV_SEQUENTIAL 0
#endif
#ifndef POSIX_MADV_RANDOM
#define POSIX_MADV_RANDOM 0
#endif
#ifndef POSIX_MADV_WILLNEED
#define POSIX_MADV_WILLNEED 0
#endif
#ifndef POSIX_MADV_DONTNEED
#define POSIX_MADV_DONTNEED 0
#endif
#ifndef O_ACCMODE
#define O_ACCMODE 0
#endif
#ifndef O_RDONLY
#define O_RDONLY 0
#endif
#ifndef O_WRONLY
#define O_WRONLY 0
#endif
#ifndef O_RDWR
#define O_RDWR 0
#endif
#ifndef O_CREAT
#define O_CREAT 0
#endif
#ifndef O_TRUNC
#define O_TRUNC 0
#endif
#ifndef O_EXCL
#define O_EXCL 0
#endif
#ifndef O_APPEND
#define O_APPEND 0
#endif
#ifndef O_ASYNC
#define O_ASYNC 0
#endif
#ifndef O_DIRECT
#define O_DIRECT 0
#endif
#ifndef O_NOATIME
#define O_NOATIME 0
#endif
#ifndef O_CLOEXEC
#define O_CLOEXEC 0
#endif
#ifndef O_NOCTTY
#define O_NOCTTY 0
#endif
#ifndef O_NOFOLLOW
#define O_NOFOLLOW 0
#endif
#ifndef O_NONBLOCK
#define O_NONBLOCK 0
#endif
#ifndef O_EXEC
#define O_EXEC 0
#endif
#ifndef O_SEARCH
#define O_SEARCH 0
#endif
#ifndef O_DIRECTORY
#define O_DIRECTORY 0
#endif
#ifndef O_DSYNC
#define O_DSYNC 0
#endif
#ifndef O_RSYNC
#define O_RSYNC 0
#endif
#ifndef O_SYNC
#define O_SYNC 0
#endif
#ifndef O_PATH
#define O_PATH 0
#endif
#ifndef O_TMPFILE
#define O_TMPFILE 0
#endif
#ifndef O_TTY_INIT
#define O_TTY_INIT 0
#endif
#ifndef S_IFIFO
#define S_IFIFO 0
#endif
#ifndef S_IFCHR
#define S_IFCHR 0
#endif
#ifndef S_IFBLK
#define S_IFBLK 0
#endif
#ifndef S_IFLNK
#define S_IFLNK 0
#endif
#ifndef S_IFREG
#define S_IFREG 0
#endif
#ifndef S_IFDIR
#define S_IFDIR 0
#endif
#ifndef S_IFWHT
#define S_IFWHT 0
#endif
#ifndef S_IFSOCK
#define S_IFSOCK 0
#endif
#ifndef S_IFMT
#define S_IFMT 0
#endif
#ifndef ST_RDONLY
#define ST_RDONLY 0
#endif
#ifndef ST_NOSUID
#define ST_NOSUID 0
#endif
#ifndef ST_NODEV
#define ST_NODEV 0
#endif
#ifndef ST_NOEXEC
#define ST_NOEXEC 0
#endif
#ifndef ST_SYNCHRONOUS
#define ST_SYNCHRONOUS 0
#endif
#ifndef ST_MANDLOCK
#define ST_MANDLOCK 0
#endif
#ifndef ST_WRITE
#define ST_WRITE 0
#endif
#ifndef ST_APPEND
#define ST_APPEND 0
#endif
#ifndef ST_IMMUTABLE
#define ST_IMMUTABLE 0
#endif
#ifndef ST_NOATIME
#define ST_NOATIME 0
#endif
#ifndef ST_NODIRATIME
#define ST_NODIRATIME 0
#endif
#ifndef ST_RELATIME
#define ST_RELATIME 0
#endif
#ifndef PROT_NONE
#define PROT_NONE 0
#endif
#ifndef PROT_EXEC
#define PROT_EXEC 0
#endif
#ifndef PROT_READ
#define PROT_READ 0
#endif
#ifndef PROT_WRITE
#define PROT_WRITE 0
#endif
#ifndef MAP_PRIVATE
#define MAP_PRIVATE 0
#endif
#ifndef MAP_SHARED
#define MAP_SHARED 0
#endif
#ifndef MAP_FIXED
#define MAP_FIXED 0
#endif
#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS 0
#endif
#ifndef MAP_LOCKED
#define MAP_LOCKED 0
#endif
#ifndef MAP_NORESERVE
#define MAP_NORESERVE 0
#endif
#ifndef MAP_POPULATE
#define MAP_POPULATE 0
#endif
#ifndef MAP_NONBLOCK
#define MAP_NONBLOCK 0
#endif
#ifndef MAP_GROWSDOWN
#define MAP_GROWSDOWN 0
#endif
#ifndef MAP_32BIT
#define MAP_32BIT 0
#endif
#ifndef MAP_HUGETLB
#define MAP_HUGETLB 0
#endif
#ifndef MAP_STACK
#define MAP_STACK 0
#endif
#ifndef MREMAP_MAYMOVE
#define MREMAP_MAYMOVE 0
#endif
#ifndef MREMAP_FIXED
#define MREMAP_FIXED 0
#endif
#ifndef F_DUPFD_CLOEXEC
#define F_DUPFD_CLOEXEC 0
#endif
#ifndef MSG_CMSG_CLOEXEC
#define MSG_CMSG_CLOEXEC 0
#endif
#ifndef SOCK_CLOEXEC
#define SOCK_CLOEXEC 0
#endif
#ifndef F_OFD_GETLK
#define F_OFD_GETLK 0
#endif
#ifndef F_OFD_SETLK
#define F_OFD_SETLK 0
#endif
#ifndef F_OFD_GETLKW
#define F_OFD_GETLKW 0
#endif
#ifndef FIFREEZE
#define FIFREEZE 0
#endif
#ifndef FITHAW
#define FITHAW 0
#endif
#ifndef FITRIM
#define FITRIM 0
#endif
#ifndef FICLONE
#define FICLONE 0
#endif
#ifndef FICLONERANGE
#define FICLONERANGE 0
#endif
#ifndef FIDEDUPERANGE
#define FIDEDUPERANGE 0
#endif
#ifndef FS_IOC_GETFLAGS
#define FS_IOC_GETFLAGS 0
#endif
#ifndef FS_IOC_SETFLAGS
#define FS_IOC_SETFLAGS 0
#endif
#ifndef FS_IOC_GETVERSION
#define FS_IOC_GETVERSION 0
#endif
#ifndef FS_IOC_SETVERSION
#define FS_IOC_SETVERSION 0
#endif
#ifndef FS_IOC_FIEMAP
#define FS_IOC_FIEMAP 0
#endif
#ifndef FS_IOC_FSGETXATTR
#define FS_IOC_FSGETXATTR 0
#endif
#ifndef FS_IOC_FSSETXATTR
#define FS_IOC_FSSETXATTR 0
#endif
#ifndef FS_IOC_SET_ENCRYPTION_POLICY
#define FS_IOC_SET_ENCRYPTION_POLICY 0
#endif
#ifndef FS_IOC_GET_ENCRYPTION_PWSALT
#define FS_IOC_GET_ENCRYPTION_PWSALT 0
#endif
#ifndef FS_IOC_GET_ENCRYPTION_POLICY
#define FS_IOC_GET_ENCRYPTION_POLICY 0
#endif
#ifndef FS_KEY_DESCRIPTOR_SIZE
#define FS_KEY_DESCRIPTOR_SIZE 0
#endif
#ifndef FS_SECRM_FL
#define FS_SECRM_FL 0
#endif
#ifndef FS_UNRM_FL
#define FS_UNRM_FL 0
#endif
#ifndef FS_COMPR_FL
#define FS_COMPR_FL 0
#endif
#ifndef FS_SYNC_FL
#define FS_SYNC_FL 0
#endif
#ifndef FS_IMMUTABLE_FL
#define FS_IMMUTABLE_FL 0
#endif
#ifndef FS_APPEND_FL
#define FS_APPEND_FL 0
#endif
#ifndef FS_NODUMP_FL
#define FS_NODUMP_FL 0
#endif
#ifndef FS_NOATIME_FL
#define FS_NOATIME_FL 0
#endif
#ifndef FS_DIRTY_FL
#define FS_DIRTY_FL 0
#endif
#ifndef FS_COMPRBLK_FL
#define FS_COMPRBLK_FL 0
#endif
#ifndef FS_NOCOMP_FL
#define FS_NOCOMP_FL 0
#endif
#ifndef FS_ENCRYPT_FL
#define FS_ENCRYPT_FL 0
#endif
#ifndef FS_BTREE_FL
#define FS_BTREE_FL 0
#endif
#ifndef FS_INDEX_FL
#define FS_INDEX_FL 0
#endif
#ifndef FS_JOURNAL_DATA_FL
#define FS_JOURNAL_DATA_FL 0
#endif
#ifndef FS_NOTAIL_FL
#define FS_NOTAIL_FL 0
#endif
#ifndef FS_DIRSYNC_FL
#define FS_DIRSYNC_FL 0
#endif
#ifndef FS_TOPDIR_FL
#define FS_TOPDIR_FL 0
#endif
#ifndef FS_FL_USER_MODIFIABLE
#define FS_FL_USER_MODIFIABLE 0
#endif
#ifndef FS_XFLAG_REALTIME
#define FS_XFLAG_REALTIME 0
#endif
#ifndef FS_XFLAG_PREALLOC
#define FS_XFLAG_PREALLOC 0
#endif
#ifndef FS_XFLAG_IMMUTABLE
#define FS_XFLAG_IMMUTABLE 0
#endif
#ifndef FS_XFLAG_APPEND
#define FS_XFLAG_APPEND 0
#endif
#ifndef FS_XFLAG_SYNC
#define FS_XFLAG_SYNC 0
#endif
#ifndef FS_XFLAG_NOATIME
#define FS_XFLAG_NOATIME 0
#endif
#ifndef FS_XFLAG_NODUMP
#define FS_XFLAG_NODUMP 0
#endif
#ifndef FS_XFLAG_RTINHERIT
#define FS_XFLAG_RTINHERIT 0
#endif
#ifndef FS_XFLAG_PROJINHERIT
#define FS_XFLAG_PROJINHERIT 0
#endif
#ifndef FS_XFLAG_NOSYMLINKS
#define FS_XFLAG_NOSYMLINKS 0
#endif
#ifndef FS_XFLAG_EXTSIZE
#define FS_XFLAG_EXTSIZE 0
#endif
#ifndef FS_XFLAG_EXTSZINHERIT
#define FS_XFLAG_EXTSZINHERIT 0
#endif
#ifndef FS_XFLAG_NODEFRAG
#define FS_XFLAG_NODEFRAG 0
#endif
#ifndef FS_XFLAG_FILESTREAM
#define FS_XFLAG_FILESTREAM 0
#endif
#ifndef FS_XFLAG_DAX
#define FS_XFLAG_DAX 0
#endif
#ifndef FS_XFLAG_HASATTR
#define FS_XFLAG_HASATTR 0
#endif
#ifndef FIEMAP_FLAG_SYNC
#define FIEMAP_FLAG_SYNC 0
#endif
#ifndef FIEMAP_FLAG_XATTR
#define FIEMAP_FLAG_XATTR 0
#endif
#ifndef FIEMAP_FLAGS_COMPAT
#define FIEMAP_FLAGS_COMPAT 0
#endif
#ifndef FIEMAP_EXTENT_LAST
#define FIEMAP_EXTENT_LAST 0
#endif
#ifndef FIEMAP_EXTENT_UNKNOWN
#define FIEMAP_EXTENT_UNKNOWN 0
#endif
#ifndef FIEMAP_EXTENT_DELALLOC
#define FIEMAP_EXTENT_DELALLOC 0
#endif
#ifndef FIEMAP_EXTENT_ENCODED
#define FIEMAP_EXTENT_ENCODED 0
#endif
#ifndef FIEMAP_EXTENT_DATA_ENCRYPTED
#define FIEMAP_EXTENT_DATA_ENCRYPTED 0
#endif
#ifndef FIEMAP_EXTENT_NOT_ALIGNED
#define FIEMAP_EXTENT_NOT_ALIGNED 0
#endif
#ifndef FIEMAP_EXTENT_DATA_INLINE
#define FIEMAP_EXTENT_DATA_INLINE 0
#endif
#ifndef FIEMAP_EXTENT_DATA_TAIL
#define FIEMAP_EXTENT_DATA_TAIL 0
#endif
#ifndef FIEMAP_EXTENT_UNWRITTEN
#define FIEMAP_EXTENT_UNWRITTEN 0
#endif
#ifndef FIEMAP_EXTENT_MERGED
#define FIEMAP_EXTENT_MERGED 0
#endif
#ifndef FIEMAP_EXTENT_SHARED
#define FIEMAP_EXTENT_SHARED 0
#endif
#ifndef SPLICE_F_MOVE
#define SPLICE_F_MOVE 0
#endif
#ifndef SPLICE_F_NONBLOCK
#define SPLICE_F_NONBLOCK 0
#endif
#ifndef SPLICE_F_MORE
#define SPLICE_F_MORE 0
#endif
#ifndef SPLICE_F_GIFT
#define SPLICE_F_GIFT 0
#endif
#ifndef EFD_CLOEXEC
#define EFD_CLOEXEC 0
#endif
#ifndef EFD_NONBLOCK
#define EFD_NONBLOCK 0
#endif
#ifndef EFD_SEMAPHORE
#define EFD_SEMAPHORE 0
#endif
#ifndef CLOCK_REALTIME
#define CLOCK_REALTIME 0
#endif
#ifndef CLOCK_MONOTONIC
#define CLOCK_MONOTONIC 0
#endif
#ifndef CLOCK_BOOTTIME
#define CLOCK_BOOTTIME 0
#endif
#ifndef CLOCK_REALTIME_ALARM
#define CLOCK_REALTIME_ALARM 0
#endif
#ifndef CLOCK_BOOTTIME_ALARM
#define CLOCK_BOOTTIME_ALARM 0
#endif
#ifndef TFD_NONBLOCK
#define TFD_NONBLOCK 0
#endif
#ifndef TFD_CLOEXEC
#define TFD_CLOEXEC 0
#endif
#ifndef TFD_TIMER_ABSTIME
#define TFD_TIMER_ABSTIME 0
#endif
#ifndef TFD_TIMER_CANCEL_ON_SET
#define TFD_TIMER_CANCEL_ON_SET 0
#endif
